@@ -25,6 +25,10 @@ clogger = logging.getLogger("dnstap_receiver.console")
 from dnstap_receiver.codecs import dnstap_pb2 
 from dnstap_receiver.codecs import fstrm 
 
+# import all inputs
+from dnstap_receiver.inputs import input_unixsocket
+from dnstap_receiver.inputs import input_tcpsocket
+
 # import all outputs
 from dnstap_receiver.outputs import output_stdout
 from dnstap_receiver.outputs import output_file
@@ -34,6 +38,7 @@ from dnstap_receiver.outputs import output_metrics
 
 from dnstap_receiver import api_server
 from dnstap_receiver import statistics
+from dnstap_receiver import dnspython_patch
 
 class UnknownValue:
     name = "-"
@@ -42,76 +47,31 @@ DNSTAP_TYPE = dnstap_pb2._MESSAGE_TYPE.values_by_number
 DNSTAP_FAMILY = dnstap_pb2._SOCKETFAMILY.values_by_number
 DNSTAP_PROTO = dnstap_pb2._SOCKETPROTOCOL.values_by_number  
 
+DFLT_LISTEN_IP = "0.0.0.0"
+DFLT_LISTEN_PORT = 6000
+
 # command line arguments definition
 parser = argparse.ArgumentParser()
 parser.add_argument("-l", 
                     help="IP of the dnsptap server to receive dnstap payloads (default: %(default)r)",
-                    default="0.0.0.0")
+                    default=DFLT_LISTEN_IP)
 parser.add_argument("-p", type=int,
                     help="Port the dnstap receiver is listening on (default: %(default)r)",
-                    default=6000)               
+                    default=DFLT_LISTEN_PORT)               
 parser.add_argument("-u", help="read dnstap payloads from unix socket")
 parser.add_argument('-v', action='store_true', help="verbose mode")   
 parser.add_argument("-c", help="external config file")   
 
-import dns.exception
-import dns.opcode
-import dns.flags
-
-# waiting fix with dnspython 2.1
-# will be removed in the future
-class _WireReader(dns.message._WireReader):
-    def read(self):
-        """issue fixed - waiting fix with dnspython 2.1"""
-        if self.parser.remaining() < 12:
-            raise dns.message.ShortHeader
-        (id, flags, qcount, ancount, aucount, adcount) = \
-            self.parser.get_struct('!HHHHHH')
-        factory = dns.message._message_factory_from_opcode(dns.opcode.from_flags(flags))
-        self.message = factory(id=id)
-        self.message.flags = flags
-        self.initialize_message(self.message)
-        self.one_rr_per_rrset = \
-            self.message._get_one_rr_per_rrset(self.one_rr_per_rrset)
-        self._get_question(dns.message.MessageSection.QUESTION, qcount)
-        
-        return self.message
-
-# waiting fix with dnspython 2.1
-# will be removed in the future
-def from_wire(wire, question_only=True):
-    """decode wire message - waiting fix with dnspython 2.1"""
-    raise_on_truncation=False
-    def initialize_message(message):
-        message.request_mac = b''
-        message.xfr = False
-        message.origin = None
-        message.tsig_ctx = None
-
-    reader = _WireReader(wire, initialize_message, question_only=question_only,
-                 one_rr_per_rrset=False, ignore_trailing=False,
-                 keyring=None, multi=False)
-    try:
-        m = reader.read()
-    except dns.exception.FormError:
-        if reader.message and (reader.message.flags & dns.flags.TC) and \
-           raise_on_truncation:
-            raise dns.message.Truncated(message=reader.message)
-        else:
-            raise
-    # Reading a truncated message might not have any errors, so we
-    # have to do this check here too.
-    if m.flags & dns.flags.TC and raise_on_truncation:
-        raise dns.message.Truncated(message=m)
-
-    return m
-    
-async def cb_ondnstap(dnstap_decoder, payload, cfg, queue, stats, geoip_reader):
+async def cb_ondnstap(dnstap_decoder, payload, cfg, queues_list, stats, geoip_reader):
     """on dnstap"""
     # decode binary payload
     dnstap_decoder.ParseFromString(payload)
     dm = dnstap_decoder.message
     
+    if cfg["trace"]["dnstap"]:
+        dns_pkt = dm.query_message if (dm.type % 2 ) == 1 else dm.response_message
+        clogger.debug("%s\n%s\n\n" % (dm,dns.message.from_wire(dns_pkt)) )
+
     # filtering by dnstap identity ?
     tap_ident = dnstap_decoder.identity.decode()
     if not len(tap_ident):
@@ -121,7 +81,7 @@ async def cb_ondnstap(dnstap_decoder, payload, cfg, queue, stats, geoip_reader):
             del dm
             return
             
-    tap = { "identity": tap_ident, "query-name": UnknownValue.name, 
+    tap = { "identity": tap_ident, "qname": UnknownValue.name, "rrtype": UnknownValue.name, 
             "query-type": UnknownValue.name, "source-ip": UnknownValue.name}
     
     # decode type message
@@ -146,31 +106,40 @@ async def cb_ondnstap(dnstap_decoder, payload, cfg, queue, stats, geoip_reader):
     # todo catching dns.message.ShortHeader exception
     # can occured with coredns if the full argument is missing
     if (dm.type % 2 ) == 1 :
-        dnstap_parsed = from_wire(dm.query_message,
-                                  question_only=True)
+        dnstap_parsed = dnspython_patch.from_wire(dm.query_message,
+                                  question_only=True)                 
         tap["length"] = len(dm.query_message)
         d1 = dm.query_time_sec +  (round(dm.query_time_nsec ) / 1000000000)
         tap["timestamp"] = datetime.fromtimestamp(d1, tz=timezone.utc).isoformat()
         tap["type"] = "query"
+        latency = 0.0
         
     # handle response message
     if (dm.type % 2 ) == 0 :
-        dnstap_parsed = from_wire(dm.response_message,
+        dnstap_parsed = dnspython_patch.from_wire(dm.response_message,
                                   question_only=True)
         tap["length"] = len(dm.response_message)
         d2 = dm.response_time_sec + (round(dm.response_time_nsec ) / 1000000000) 
         tap["timestamp"] = datetime.fromtimestamp(d2, tz=timezone.utc).isoformat()
         tap["type"] = "response"
+
+        # compute latency 
+        d1 = dm.query_time_sec +  (round(dm.query_time_nsec ) / 1000000000)
+        latency = round(d2-d1,3)
+    
+    tap["latency"] = latency
         
     # common params
     if len(dnstap_parsed.question):
         tap["qname"] = dnstap_parsed.question[0].name.to_text()
         tap["rrtype"] = dns.rdatatype.to_text(dnstap_parsed.question[0].rdtype)
     tap["rcode"] = dns.rcode.to_text(dnstap_parsed.rcode())
-    
+    tap["id"] = dnstap_parsed.id
+    tap["flags"] = dns.flags.to_text(dnstap_parsed.flags)
+
     # filtering by qname ?
     if cfg["filter"]["qname-regex"] is not None:
-        if re.match(cfg["filter"]["qname-regex"], tap["query-name"]) is None:
+        if re.match(cfg["filter"]["qname-regex"], tap["qname"]) is None:
             del dm; del tap;
             return
 
@@ -178,7 +147,10 @@ async def cb_ondnstap(dnstap_decoder, payload, cfg, queue, stats, geoip_reader):
     if geoip_reader is not None:
         try:
             response = geoip_reader.city(tap["source-ip"])
-            tap["country"] = response.country.name
+            if cfg["geoip"]["country-iso"]:
+                tap["country"] = response.country.iso_code
+            else:
+                tap["country"] = response.country.name
             if response.city.name is not None:
                 tap["city"] = response.city.name
             else:
@@ -191,9 +163,10 @@ async def cb_ondnstap(dnstap_decoder, payload, cfg, queue, stats, geoip_reader):
     stats.record(tap=tap)
         
     # append the dnstap message to the queue
-    queue.put_nowait(tap)
-
-async def cb_onconnect(reader, writer, cfg, queue, stats, geoip_reader):
+    for q in queues_list:
+        q.put_nowait(tap)
+    
+async def cb_onconnect(reader, writer, cfg, queues_list, stats, geoip_reader):
     """callback when a connection is established"""
     # get peer name
     peername = writer.get_extra_info('peername')
@@ -245,7 +218,7 @@ async def cb_onconnect(reader, writer, cfg, queue, stats, geoip_reader):
 
                 # handle the DATA frame
                 if fs == fstrm.FSTRM_DATA_FRAME:
-                    loop.create_task(cb_ondnstap(dnstap_decoder, payload, cfg, queue, stats, geoip_reader))
+                    loop.create_task(cb_ondnstap(dnstap_decoder, payload, cfg, queues_list, stats, geoip_reader))
                     
                 # handle the control frame READY
                 if fs == fstrm.FSTRM_CONTROL_READY:
@@ -293,12 +266,6 @@ def setup_config(args):
     except yaml.parser.ParserError:
         print("invalid default yaml config file")
         sys.exit(1)
-    
-    # update default config with command line arguments
-    cfg["trace"]["verbose"] = args.v
-    cfg["input"]["unix-socket"]["path"] = args.u
-    cfg["input"]["tcp-socket"]["local-address"] = args.l
-    cfg["input"]["tcp-socket"]["local-port"] = args.p
 
     # overwrite config with external file ?    
     if args.c:
@@ -311,6 +278,17 @@ def setup_config(args):
         except yaml.parser.ParserError:
             print("external invalid yaml config file")
             sys.exit(1)
+
+    # update default config with command line arguments
+    if args.v:
+        cfg["trace"]["verbose"] = args.v    
+    if args.u is not None:
+        cfg["input"]["unix-socket"]["path"] = args.u
+    if args.l != DFLT_LISTEN_IP:
+        cfg["input"]["tcp-socket"]["local-address"] = args.l
+    if args.l != DFLT_LISTEN_PORT:
+        cfg["input"]["tcp-socket"]["local-port"] = args.p
+
     return cfg
     
 def setup_logger(cfg):
@@ -331,67 +309,66 @@ def setup_logger(cfg):
     
     clogger.addHandler(lh)
     
-def setup_outputs(cfg, queue, stats, loop):
+def setup_outputs(cfg, stats, loop):
     """setup outputs"""
     conf = cfg["output"]
 
+    queues_list = []
     if conf["syslog"]["enable"]:
         if not output_syslog.checking_conf(cfg=conf["syslog"]): return
-        loop.create_task(output_syslog.handle(conf["syslog"], queue, stats))    
+        queue_syslog = asyncio.Queue()
+        queues_list.append(queue_syslog)
+        loop.create_task(output_syslog.handle(conf["syslog"], queue_syslog, stats))    
 
     if conf["tcp-socket"]["enable"]:
         if not output_tcp.checking_conf(cfg=conf["tcp-socket"]): return
-        loop.create_task(output_tcp.handle(conf["tcp-socket"], queue, stats))
+        queue_tcpsocket = asyncio.Queue()
+        queues_list.append(queue_tcpsocket)
+        loop.create_task(output_tcp.handle(conf["tcp-socket"], queue_tcpsocket, stats))
                                                
     if conf["file"]["enable"]:
         if not output_file.checking_conf(cfg=conf["file"]): return
-        loop.create_task(output_file.handle(conf["file"], queue, stats))
+        queue_file = asyncio.Queue()
+        queues_list.append(queue_file)
+        loop.create_task(output_file.handle(conf["file"], queue_file, stats))
                                               
     if conf["stdout"]["enable"]:
         if not output_stdout.checking_conf(cfg=conf["stdout"]): return
-        loop.create_task(output_stdout.handle(conf["stdout"], queue, stats))
+        queue_stdout = asyncio.Queue()
+        queues_list.append(queue_stdout)
+        loop.create_task(output_stdout.handle(conf["stdout"], queue_stdout, stats))
 
     if conf["metrics"]["enable"]:
         if not output_metrics.checking_conf(cfg=conf["metrics"]): return
-        loop.create_task(output_metrics.handle(conf["metrics"], queue, stats))
+        queue_metrics = asyncio.Queue()
+        queues_list.append(queue_metrics)
+        loop.create_task(output_metrics.handle(conf["metrics"], queue_metrics, stats))
 
-def setup_inputs(args, cfg, queue, stats, loop, geoip_reader):
+    return queues_list
+    
+def setup_inputs(args, cfg, queues_list, stats, loop, geoip_reader):
     """setup inputs"""
+    # define callback on new connection
+    cb_lambda = lambda r, w: cb_onconnect(r, w, cfg, queues_list, stats, geoip_reader)
+    
     # asynchronous unix socket
     if cfg["input"]["unix-socket"]["path"] is not None:
-        clogger.debug("Input handler: unix socket")
-        clogger.debug("Input handler: listening on %s" % args.u)
-        socket_server = asyncio.start_unix_server(lambda r, w: cb_onconnect(r, w, cfg, queue, stats, geoip_reader),
-                                                  path=cfg["input"]["unix-socket"]["path"],
-                                                  loop=loop)
+        socket_server = input_unixsocket.start_input(cfg=cfg["input"]["unix-socket"], cb_onconnect=cb_lambda, loop=loop)
+
     # default mode: asynchronous tcp socket
     else:
-        clogger.debug("Input handler: tcp socket")
-        
-        ssl_context = None
-        if cfg["input"]["tcp-socket"]["tls-support"]:
-            ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
-            ssl_context.load_cert_chain(certfile=cfg["input"]["tcp-socket"]["tls-server-cert"], 
-                                        keyfile=cfg["input"]["tcp-socket"]["tls-server-key"])
-            clogger.debug("Input handler - tls support enabled")
-        clogger.debug("Input handler: listening on %s:%s" % (cfg["input"]["tcp-socket"]["local-address"],
-                                              cfg["input"]["tcp-socket"]["local-port"])), 
-        socket_server = asyncio.start_server(lambda r, w: cb_onconnect(r, w, cfg, queue, stats, geoip_reader),
-                                             cfg["input"]["tcp-socket"]["local-address"],
-                                             cfg["input"]["tcp-socket"]["local-port"],
-                                             ssl=ssl_context,
-                                             loop=loop)
+        socket_server = input_tcpsocket.start_input(cfg=cfg["input"]["tcp-socket"], cb_onconnect=cb_lambda, loop=loop)
                                              
     # run until complete
     loop.run_until_complete(socket_server)
     
-def setup_api(cfg, queue, stats, loop):
+def setup_api(cfg, stats, loop):
     """setup web api"""
     if cfg["web-api"]["enable"]:
         api_svr = api_server.create_server(loop, cfg=cfg["web-api"], 
                                            stats=stats, cfg_stats=cfg["statistics"])
         loop.run_until_complete(api_svr)
-    
+
 def setup_geoip(cfg):
     if not cfg["enable"]: return None
     if cfg["city-database"] is None: return None
@@ -409,13 +386,13 @@ def start_receiver():
     # Handle command-line arguments.
     args = parser.parse_args()
 
-    # init config
+    # setup config
     cfg = setup_config(args=args)
-            
-    # init logging
+  
+    # setup logging
     setup_logger(cfg=cfg["trace"])
 
-    # setup geoip is activated 
+    # setup geoip if enabled 
     geoip_reader = setup_geoip(cfg=cfg["geoip"])
     
     # add debug message if external config is used
@@ -424,18 +401,17 @@ def start_receiver():
     # start receiver and get event loop
     clogger.debug("Start receiver...")
     loop = asyncio.get_event_loop()
-    queue = asyncio.Queue()
-    stats = statistics.Statistics()
+    stats = statistics.Statistics(cfg=cfg["statistics"])
     loop.create_task(statistics.watcher(stats))
     
     # prepare outputs
-    setup_outputs(cfg, queue, stats, loop)
+    queues_list = setup_outputs(cfg, stats, loop)
     
     # prepare inputs
-    setup_inputs(args, cfg, queue, stats, loop, geoip_reader)
+    setup_inputs(args, cfg, queues_list, stats, loop, geoip_reader)
 
-    # start the rest api
-    setup_api(cfg, queue, stats, loop)
+    # start the http api
+    setup_api(cfg, stats, loop)
 
     # run event loop 
     try:
